import pytest
from dotenv import load_dotenv

load_dotenv()


@pytest.fixture()
def storage_options():
    return {"use_emulator": "True"}


@pytest.fixture(scope="session", autouse=True)
def spawn_azurite():
    import test_server
    import os

    if os.getenv("NO_AZURITE_DOCKER", "0") == "1":
        test_server.upload_to_azurite()
        yield None
    else:
        azurite = test_server.start_azurite()
        yield azurite
        if (
            os.getenv("KEEP_AZURITE_DOCKER", "0") == "0"
        ):  # can be handy during development
            azurite.stop()



import duckdb

res = duckdb.execute(
    r"""select name from parquet_schema('C:\Projects\BMS_Github\deltalake2db\tests\data\faker2\Ec\*')"""
)
print([c[0] for c in res.description])
print(res.fetch_df())



from pathlib import Path
from typing import TYPE_CHECKING, cast, Union, Optional, Callable, overload

from deltalake2db.azure_helper import (
    get_storage_options_object_store,
    get_storage_options_fsspec,
)
from deltalake2db.filter_by_meta import _can_filter

if TYPE_CHECKING:
    import polars as pl
    import polars.schema as pl_schema
    from azure.core.credentials import TokenCredential
from collections import OrderedDict
import os
from deltalake2db.delta_meta_retrieval import (
    PolarsEngine,
    get_meta,
    DataType,
    field_to_type,
    Field,
)


class PolarsSettings:
    use_pyarrow = False
    exclude_fields: Optional[list[str]] = None
    fields: Optional[list[str]] = None

    timestamp_ntz_type: "pl.Datetime"
    timestamp_type: "pl.Datetime"

    def __init__(
        self,
        *,
        timestamp_ntz_type: "Optional[pl.Datetime]" = None,
        timestamp_type: "Optional[pl.Datetime]" = None,
        exclude_fields: Optional[list[str]] = None,
        fields: Optional[list[str]] = None,
        use_pyarrow=False,
    ):
        import polars as pl

        self.timestamp_ntz_type = timestamp_ntz_type or pl.Datetime(
            time_unit="us", time_zone=None
        )
        self.timestamp_type = timestamp_type or pl.Datetime(
            time_unit="us", time_zone="utc"
        )
        self.exclude_fields = exclude_fields
        self.fields = fields
        self.use_pyarrow = use_pyarrow


def _get_expr(
    base_expr: "Union[pl.Expr,None]",
    dtype: "DataType",
    meta: "Optional[Field]",
    parquet_field: "Optional[pl.PolarsDataType]",
    settings: "PolarsSettings",
) -> "pl.Expr":
    import polars as pl

    if parquet_field is None:
        return (
            pl.lit(None, _get_type(dtype, False, settings)).alias(meta["name"])
            if meta
            else pl.lit(None, _get_type(dtype, False, settings))
        )

    pn = (
        meta.get("metadata", {}).get("delta.columnMapping.physicalName", meta["name"])
        if meta
        else None
    )
    if base_expr is None:
        assert pn is not None
        base_expr = pl.col(pn)
    if isinstance(dtype, dict) and dtype["type"] == "struct":

        def _get_sub_type(name: str) -> "pl.PolarsDataType| None":
            if parquet_field is None:
                return None
            assert isinstance(parquet_field, pl.Struct)
            for f in parquet_field.fields:
                if f.name == name:
                    return f.dtype
            return None

        struct_vl = (
            pl.when(base_expr.is_null())
            .then(pl.lit(None))
            .otherwise(
                pl.struct(
                    *[
                        _get_expr(
                            base_expr.struct.field(
                                subfield.get("metadata", {}).get(
                                    "delta.columnMapping.physicalName", subfield["name"]
                                )
                            ),
                            field_to_type(subfield),
                            subfield,
                            _get_sub_type(
                                subfield.get("metadata", {}).get(
                                    "delta.columnMapping.physicalName", subfield["name"]
                                )
                            ),
                            settings,
                        )
                        for subfield in dtype["fields"]
                    ]
                )
            )
        )
        return struct_vl.alias(meta["name"]) if meta else struct_vl
    elif isinstance(dtype, dict) and dtype["type"] == "array":
        list_vl = (
            pl.when(base_expr.is_null())
            .then(pl.lit(None))
            .otherwise(
                base_expr.list.eval(
                    _get_expr(
                        pl.element(),
                        dtype=dtype["elementType"],
                        meta=meta,
                        parquet_field=cast(pl.List, parquet_field).inner,
                        settings=settings,
                    )
                )
            )
        )
        return list_vl.alias(meta["name"]) if meta else list_vl
    return base_expr.alias(meta["name"]) if meta else base_expr


def _get_type(
    dtype: "DataType", physical: bool, settings: "PolarsSettings"
) -> "Union[pl.DataType, type[pl.DataType]]":
    import polars as pl

    if isinstance(dtype, dict) and dtype["type"] == "struct":
        return pl.Struct(
            [
                pl.Field(
                    f["name"]
                    if not physical
                    else f.get("metadata", {}).get(
                        "delta.columnMapping.physicalName", f["name"]
                    ),
                    _get_type(field_to_type(f), physical, settings),
                )
                for f in dtype["fields"]
            ]
        )
    if isinstance(dtype, dict) and dtype["type"] == "array":
        return pl.List(_get_type(dtype["elementType"], physical, settings))
    if isinstance(dtype, dict) and dtype["type"] == "map":
        return pl.List(
            pl.Struct(
                [
                    pl.Field(
                        "key", _get_type(dtype["keyType"], physical, settings=settings)
                    ),
                    pl.Field(
                        "value",
                        _get_type(dtype["valueType"], physical, settings=settings),
                    ),
                ]
            )
        )  # Polars models maps as structs with "key" and "value" fields

    dtype_str = dtype
    if dtype_str == "string":
        return pl.String
    elif dtype_str in ["int", "integer"]:
        return pl.Int32
    elif dtype_str == "long":
        return pl.Int64
    elif dtype_str == "double":
        return pl.Float64
    elif dtype_str == "float":
        return pl.Float32
    elif dtype_str == "boolean":
        return pl.Boolean
    elif dtype_str == "date":
        return pl.Date
    elif dtype_str == "timestamp":
        return settings.timestamp_type
    elif dtype_str == "binary":
        return pl.Binary
    elif dtype_str.startswith("decimal"):
        precision, scale = dtype_str.split("(")[1].split(")")[0].split(",")
        return pl.Decimal(int(precision), int(scale))
    elif dtype_str == "short":
        return pl.Int16
    elif dtype == "byte":
        return pl.Int8
    elif dtype_str == "null":
        return pl.Null
    elif dtype_str in ["timestampNtz", "timestamp_ntz"]:
        return settings.timestamp_ntz_type
    else:
        return pl.Object


def get_polars_schema(
    delta_table: "Union[Path, str]",
    physical_name: bool = False,
    settings: "Optional[PolarsSettings]" = None,
    storage_options: "Optional[dict]" = None,
    version: "Optional[int]" = None,
) -> "pl.Schema":
    from .protocol_check import check_is_supported
    import polars as pl

    if settings is None:
        settings = PolarsSettings()

    delta_meta = get_meta(
        PolarsEngine(storage_options), str(delta_table), version=version
    )

    check_is_supported(delta_meta)
    res_dict = OrderedDict()
    assert delta_meta.schema is not None
    for f in delta_meta.schema["fields"]:
        pn = f["name"]
        if settings.exclude_fields and f["name"] in settings.exclude_fields:
            continue
        if settings.fields and f["name"] not in settings.fields:
            continue
        if physical_name:
            pn = f.get("metadata", {}).get(
                "delta.columnMapping.physicalName", f["name"]
            )
        res_dict[pn] = _get_type(field_to_type(f), physical_name, settings)
    return pl.Schema(res_dict)


def _has_datetime_field(dt: "pl.PolarsDataType") -> bool:
    import polars as pl

    if isinstance(dt, pl.Datetime) or dt == pl.Datetime:
        return True
    if isinstance(dt, pl.Struct):
        return any(_has_datetime_field(f.dtype) for f in dt.fields)
    if isinstance(dt, pl.List):
        return _has_datetime_field(dt.inner)
    if isinstance(dt, pl.Array):
        return _has_datetime_field(dt.inner)
    return False


def _cast_schema(ds: "pl.DataFrame", schema: "pl_schema.Schema") -> "pl.DataFrame":
    import polars as pl

    exprs = []
    for name, dtype in schema.items():
        if name not in ds.columns:
            exprs.append(pl.lit(None, dtype).cast(dtype).alias(name))
        else:
            exprs.append(pl.col(name).cast(dtype).alias(name))
    return ds.select(exprs)


def _versiontuple(v):
    return tuple(map(int, (v.split("."))))


@overload
def scan_delta_union(
    delta_table: "Union[Path, str]",
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    version: "Optional[int]" = None,
) -> "pl.LazyFrame": ...


@overload
def scan_delta_union(
    delta_table: "Union[Path, str]",
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    settings: "Optional[PolarsSettings]" = None,
    version: "Optional[int]" = None,
) -> "Union[pl.LazyFrame, pl.DataFrame]": ...


def scan_delta_union(
    delta_table: "Union[Path, str]",
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    settings: "Optional[PolarsSettings]" = None,
    version: "Optional[int]" = None,
) -> "Union[pl.LazyFrame, pl.DataFrame]":
    import polars as pl
    import polars.datatypes as pldt
    from .protocol_check import check_is_supported

    if settings is None:
        settings = PolarsSettings()

    path_for_delta, storage_options_for_delta = get_storage_options_object_store(
        delta_table, storage_options, get_credential
    )
    delta_meta = get_meta(
        PolarsEngine(storage_options_for_delta), str(delta_table), version=version
    )
    check_is_supported(delta_meta)
    all_ds: Union[list[pl.LazyFrame], list[pl.DataFrame]] = []
    assert delta_meta.schema is not None
    assert delta_meta.last_metadata is not None
    all_fields = delta_meta.schema["fields"]
    physical_schema = get_polars_schema(
        delta_table,
        physical_name=True,
        settings=settings,
        storage_options=storage_options_for_delta,
        version=version,
    )
    physical_schema_no_parts = physical_schema.copy()

    logical_to_physical = {
        f["name"]: f.get("metadata", {}).get(
            "delta.columnMapping.physicalName", f["name"]
        )
        for f in all_fields
        if (settings.exclude_fields is None or f["name"] not in settings.exclude_fields)
        and (settings.fields is None or f["name"] in settings.fields)
    }
    for pc in delta_meta.last_metadata.get("partitionColumns", []):
        physical_schema_no_parts.pop(logical_to_physical.get(pc, pc))

    datetime_fields = {
        key: value
        for key, value in physical_schema_no_parts.items()
        if _has_datetime_field(value)
    }

    dt_mapping: Optional[dict[str, Union[pl.DataType, pldt.DataTypeClass]]] = None
    dt_mapping_complete = False
    if (
        settings.use_pyarrow
        and storage_options
        and path_for_delta
        and "://" in str(path_for_delta)
    ):
        import fsspec

        proto = str(path_for_delta).split("://")[0]
        if proto == "az":
            proto = "abfss"
        storage_options_for_fsspec = get_storage_options_fsspec(storage_options)
        fs = fsspec.filesystem(proto, **storage_options_for_fsspec)
        pyarrow_opts = {"filesystem": fs}  # type: ignore
    else:
        pyarrow_opts = None
        storage_options_for_fsspec = None

    for ac in delta_meta.add_actions.values():
        if conditions is not None and _can_filter(ac, conditions):
            continue
        fullpath = os.path.join(path_for_delta, ac["path"]).replace("\\", "/")
        if settings.use_pyarrow:
            ds = pl.read_parquet(
                fullpath,
                storage_options=storage_options_for_fsspec,
                glob=False,
                use_pyarrow=True,
                pyarrow_options=pyarrow_opts,
            )
        else:
            ds = None

        if datetime_fields and not dt_mapping_complete:
            # we need this as long as https://github.com/pola-rs/polars/issues/19533 is not done
            file_schema = (
                pl.scan_parquet(
                    fullpath,
                    storage_options=storage_options_for_delta,
                    glob=False,
                ).collect_schema()
                if ds is None
                else ds.schema
            )
            dt_mapping_complete = True
            for pn in datetime_fields.keys():
                dt_mapping = {}

                pys_type = file_schema.get(pn)
                if pys_type is None:
                    dt_mapping_complete = False
                    continue
                if isinstance(pys_type, pl.Int64):  # int64 is microsecond
                    dt_mapping[pn] = pl.Datetime(time_unit="us", time_zone="utc")
                else:
                    dt_mapping[pn] = pys_type

                physical_schema_no_parts[pn] = dt_mapping[pn]

        base_ds = (
            pl.scan_parquet(
                fullpath,
                storage_options=storage_options_for_delta,
                glob=False,
                schema=physical_schema_no_parts,
                missing_columns="insert",
                extra_columns="ignore",
            )
            if _versiontuple(pl.__version__) >= (1, 31)
            else pl.scan_parquet(
                fullpath,
                storage_options=storage_options_for_delta,
                glob=False,
                schema=physical_schema_no_parts,
                allow_missing_columns=True,
            )
            if ds is None
            else _cast_schema(ds, physical_schema_no_parts)
        )
        selects = []
        for field in all_fields:
            if settings.exclude_fields and field["name"] in settings.exclude_fields:
                continue
            if settings.fields and field["name"] not in settings.fields:
                continue
            pn = field.get("metadata", {}).get(
                "delta.columnMapping.physicalName", field["name"]
            )
            if "partitionValues" in ac and pn in ac["partitionValues"]:
                part_vl = ac["partitionValues"][pn]
                selects.append(
                    pl.lit(
                        part_vl, _get_type(field_to_type(field), False, settings)
                    ).alias(field["name"])
                )
            elif "partition." + pn in ac:
                part_vl = ac["partition." + pn]
                selects.append(
                    pl.lit(
                        part_vl, _get_type(field_to_type(field), False, settings)
                    ).alias(field["name"])
                )
            elif "partitionValues" in ac and field["name"] in ac["partitionValues"]:
                # as of delta 0.14
                part_vl = ac["partitionValues"][field["name"]]
                selects.append(
                    pl.lit(
                        part_vl, _get_type(field_to_type(field), False, settings)
                    ).alias(field["name"])
                )
            elif "partition." + field["name"] in ac:
                # as of delta 0.14
                part_vl = ac["partition." + field["name"]]
                selects.append(
                    pl.lit(
                        part_vl, _get_type(field_to_type(field), False, settings)
                    ).alias(field["name"])
                )
            else:
                selects.append(
                    _get_expr(
                        None,
                        field_to_type(field),
                        field,
                        physical_schema.get(
                            field.get("metadata", {}).get(
                                "delta.columnMapping.physicalName", field["name"]
                            )
                        ),
                        settings,
                    )
                )

        ds = (
            base_ds.select(*selects).filter(**conditions)
            if conditions is not None
            else base_ds.select(*selects)
        )
        all_ds.append(ds)  # type: ignore
    if len(all_ds) == 0:
        return pl.DataFrame(
            data=[],
            schema={
                f["name"]: _get_type(field_to_type(f), False, settings)
                for f in all_fields
            },
        ).lazy()
    return pl.concat(all_ds, how="diagonal_relaxed")



import sqlglot as sg
import sqlglot.expressions as ex

print(repr(ex.DataType.build("decimal(24,5)")))
print(
    repr(
        sg.parse_one(
            "select 234324::decimal(24,5) as tester",
            dialect="duckdb",
        )
    )
)



from deltalake2db.delta_meta_retrieval import get_meta, PolarsEngine
from datetime import datetime, timezone


def test_last_write_time():
    import deltalake

    for tbl in ["user"]:
        meta = get_meta(PolarsEngine(None), f"tests/data/{tbl}")
        assert meta.last_write_time is not None
        dt = deltalake.DeltaTable(f"tests/data/{tbl}")

        t2 = datetime.fromtimestamp(
            dt.history(1)[-1]["timestamp"] / 1000.0,
            tz=timezone.utc,
        )
        assert meta.last_write_time == t2



from pathlib import Path
import docker
from docker.models.containers import Container
from time import sleep
from typing import Optional, cast
import docker.errors
import os


def get_test_blobstorage():
    constr = os.getenv(
        "TEST_BLOB_CONSTR",
        "DefaultEndpointsProtocol=http;AccountName=devstoreaccount1;AccountKey=Eby8vdM02xNOcqFlqUwJPLlmEtlCDXJ1OUzFT50uSRZ6IFsuFq2UVErCz4I6tq/K1SZFPTOtr/KBHBeksoGMGw==;BlobEndpoint=http://127.0.0.1:10000/devstoreaccount1;",
    )
    from azure.storage.blob import ContainerClient

    cc = ContainerClient.from_connection_string(constr, "testlakedb")
    if not cc.exists():
        cc.create_container()
    else:
        cc.delete_container()
        cc.create_container()
    return cc


def upload_to_azurite():
    from azure.core.exceptions import ResourceExistsError

    with get_test_blobstorage() as cc:
        fakeroot = Path("tests/data/faker2")
        for root, _, fls in os.walk(fakeroot):
            for fl in fls:
                try:
                    rel = str(Path(root).relative_to(fakeroot))
                    with open(os.path.join(root, fl), "rb") as f:
                        target_path = (
                            f"td/delta/fake/{rel}/{fl}"
                            if rel != "."
                            else f"td/delta/fake/{fl}"
                        )
                        cc.upload_blob(target_path, f)
                except ResourceExistsError:
                    pass  # already uploaded


def start_azurite() -> Container:
    client = docker.from_env()  # code taken from https://github.com/fsspec/adlfs/blob/main/adlfs/tests/conftest.py#L72
    azurite_server: Optional[Container] = None
    try:
        m = cast(Container, client.containers.get("test4azurite"))
        if m.status == "running":
            upload_to_azurite()
            return m
        else:
            azurite_server = m
    except docker.errors.NotFound:
        pass

    if azurite_server is None:
        azurite_server = client.containers.run(
            "mcr.microsoft.com/azure-storage/azurite:latest",
            detach=True,
            name="test4azurite",
            ports={"10000/tcp": 10000, "10001/tcp": 10001, "10002/tcp": 10002},
        )  # type: ignore
    assert azurite_server is not None
    azurite_server.start()
    print(azurite_server.status)
    sleep(20)
    upload_to_azurite()
    print("Successfully created azurite container...")
    return azurite_server



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from deltalake2db.delta_meta_retrieval import MetaState


supported_reader_features = [
    "columnMapping",
    "timestampNtz",
    "typeWidening",
]  # not: deletionVectors, v2Checkpoint


def is_protocol_supported(dt: "MetaState") -> bool:
    prot = dt.protocol
    assert prot is not None
    if prot["minReaderVersion"] <= 3:
        return True
    assert prot["readerFeatures"] is not None
    un_supported = [
        f for f in prot["readerFeatures"] if f not in supported_reader_features
    ]
    return not any(un_supported)


class DeltaProtocolError(Exception):
    pass


def check_is_supported(dt: "MetaState") -> None:
    prot = dt.protocol
    assert prot is not None
    if prot["minReaderVersion"] <= 3:
        return
    assert prot["readerFeatures"] is not None
    un_supported = [
        f for f in prot["readerFeatures"] if f not in supported_reader_features
    ]
    if not any(un_supported):
        return
    raise DeltaProtocolError(
        f"Delta table features not supported: {', '.join(un_supported)}"
    )



from collections import OrderedDict
import duckdb
import polars as pl
import pytest


@pytest.mark.parametrize("use_delta_ext", [False, True])
def test_col_mapping(use_delta_ext):
    dt = "tests/data/faker2"

    from deltalake2db import get_sql_for_delta

    if (
        use_delta_ext
    ):  # see https://github.com/duckdb/duckdb-delta/issues/50#issuecomment-2730119786
        select = ["Super Name", "Company Very Short", "age", "new_name"]
    else:
        select = None
    with duckdb.connect() as con:
        sql = get_sql_for_delta(
            dt, duck_con=con, use_delta_ext=use_delta_ext, select=select
        )
        print(sql)
        con.execute("create view delta_table as " + sql)

        df = pl.from_arrow(con.execute("select * from delta_table").fetch_arrow_table())
        assert isinstance(df, pl.DataFrame)
        print(df)
    if not use_delta_ext:
        coord_field = df.schema["main_coord"]
        assert isinstance(coord_field, pl.Struct)
        fields = coord_field.fields
        assert "lat" in [f.name for f in fields]
        assert "lon" in [f.name for f in fields]

    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    if not use_delta_ext:
        assert df.schema == OrderedDict(
            [
                ("Super Name", pl.String),
                ("Company Very Short", pl.String),
                ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
                ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
                ("age", pl.List(pl.Int64)),
                ("new_name", pl.String),
            ]
        )
    else:
        assert df.schema == OrderedDict(
            [
                ("Super Name", pl.String),
                ("Company Very Short", pl.String),
                ("age", pl.List(pl.Int64)),
                ("new_name", pl.String),
            ]
        )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


@pytest.mark.parametrize("use_delta_ext", [False, True])
def test_strange_cols(use_delta_ext):
    dt = "tests/data/user"

    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(
            con, dt, "delta_table", use_delta_ext=use_delta_ext
        )
        con.execute("select * from delta_table")
        assert con.description is not None
        col_names = [c[0] for c in con.description]
        assert "time stämp" in col_names
        print("\n")
        print(con.df())


def test_filter_number():
    dt = "tests/data/user"

    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(con, dt, "delta_table", conditions={"Age": 23.0})
        con.execute("select FirstName from delta_table")
        assert con.description is not None
        col_names = [c[0] for c in con.description]
        assert col_names == ["FirstName"]
        names = con.fetchall()
        assert len(names) == 1
        assert names[0][0] == "Peter"
    with duckdb.connect() as con:
        duckdb_create_view_for_delta(con, dt, "delta_table_1", conditions={"Age": 23.0})
        con.execute("select * from delta_table_1")
        assert con.description is not None
        col_types_1 = [c[1] for c in con.description]
        duckdb_create_view_for_delta(
            con, dt, "delta_table_2", conditions={"Age": 500.0}
        )
        con.execute("select * from delta_table_2")
        assert con.description is not None
        col_types_2 = [c[1] for c in con.description]
        assert col_types_1 == col_types_2


def test_filter_name():
    dt = "tests/data/user"

    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(
            con, dt, "delta_table", conditions={"FirstName": "Peter"}
        )
        con.execute("select FirstName from delta_table")
        assert con.description is not None
        col_names = [c[0] for c in con.description]
        names = con.fetchall()
        assert len(names) == 1
        assert names[0][0] == "Peter"


def test_user_empty():
    dt = "tests/data/user_empty"

    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(con, dt, "delta_table")
        con.execute("select * from delta_table")
        assert con.description is not None
        col_names = [c[0] for c in con.description]
        assert "time stämp" in col_names
        assert len(con.fetchall()) == 0


def test_user_add():
    import shutil
    import pandas as pd
    from deltalake import DeltaTable

    shutil.rmtree("tests/data/_user2", ignore_errors=True)
    shutil.copytree("tests/data/user", "tests/data/_user2")

    dt = DeltaTable("tests/data/_user2")
    old_version = dt.version()
    from deltalake.writer import write_deltalake
    from deltalake import __version__

    if __version__.startswith("0."):
        engine_args = {"engine": "rust"}
    else:
        engine_args = {}
    write_deltalake(
        dt,
        pd.DataFrame({"User - iD": [1555], "FirstName": ["Hansueli"]}),
        schema_mode="merge",
        mode="append",
        **engine_args,  # type: ignore
    )
    dt.update_incremental()

    dt_o = DeltaTable("tests/data/_user2")
    dt_o.load_as_version(old_version)

    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(con, dt.table_uri, "delta_table_n")
        duckdb_create_view_for_delta(
            con, dt_o.table_uri, "delta_table_o", version=old_version
        )
        con.execute(
            'select "User - iD" from delta_table_n except select "User - iD" from delta_table_o'
        )
        res = con.fetchall()
        assert len(res) == 1
        assert res[0][0] == 1555


def test_empty_struct():
    # >>> duckdb.execute("""Select { 'lat': 1 } as tester union all select Null""").fetchall()
    import pyarrow as pa
    import pyarrow.compute as pac

    dt = "tests/data/faker2"

    from deltalake2db import get_sql_for_delta

    with duckdb.connect() as con:
        sql = get_sql_for_delta(dt, duck_con=con)
        print(sql)
        con.execute("create view delta_table as " + sql)

        df = con.execute("select * from delta_table").fetch_arrow_table()
        print(df)
        mc = (
            df.filter(pac.field("new_name") == "Hans Heiri")
            .select(["main_coord"])
            .to_pylist()
        )
        assert len(mc) == 1
        assert mc[0]["main_coord"] is None



from collections import OrderedDict
from deltalake import DeltaTable
import polars as pl
import pytest

from deltalake2db.polars import PolarsSettings


@pytest.mark.skip(reason="Would need a real account for this")
def test_chain():
    opts = {"use_emulator": "True", "chain": "default"}
    from deltalake2db import polars_scan_delta

    df = polars_scan_delta("az://testlakedb/td/delta/fake", storage_options=opts)

    df = df.collect()

    main_coord_field = df.schema["main_coord"]
    assert isinstance(main_coord_field, pl.Struct)
    fields = main_coord_field.fields
    assert "lat" in [f.name for f in fields]
    assert "lon" in [f.name for f in fields]

    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    assert df.schema == OrderedDict(
        [
            ("Super Name", pl.String),
            ("Company Very Short", pl.String),
            ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
            ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
            ("age", pl.List(pl.Int64)),
            ("new_name", pl.String),
        ]
    )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


@pytest.mark.parametrize("use_pyarrow", [True, False])
def test_col_mapping(storage_options, use_pyarrow):
    from deltalake2db import polars_scan_delta

    df = polars_scan_delta(
        "az://testlakedb/td/delta/fake",
        storage_options=storage_options,
        settings=PolarsSettings(
            use_pyarrow=use_pyarrow,
        ),
    )
    if isinstance(df, pl.LazyFrame):
        df = df.collect()
    coord_field = df.schema["main_coord"]
    assert isinstance(coord_field, pl.Struct)
    fields = coord_field.fields
    assert "lat" in [f.name for f in fields]
    assert "lon" in [f.name for f in fields]

    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    assert df.schema == OrderedDict(
        [
            ("Super Name", pl.String),
            ("Company Very Short", pl.String),
            ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
            ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
            ("age", pl.List(pl.Int64)),
            ("new_name", pl.String),
        ]
    )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


def test_empty_struct(storage_options):
    from deltalake2db import polars_scan_delta

    df = polars_scan_delta(
        "az://testlakedb/td/delta/fake", storage_options=storage_options
    )

    df = df.collect()

    mc = df.filter(new_name="Hans Heiri").select("main_coord").to_dicts()
    assert len(mc) == 1
    assert mc[0]["main_coord"] is None



from deltalake import write_deltalake
import polars as pl
import pytest
from polars.testing import assert_frame_equal
from deltalake2db import polars_scan_delta
from datetime import datetime


@pytest.fixture()
def data_batch_1():
    return pl.DataFrame(
        {
            "foo": [1, 2, 3, 4, 5, 6, 7, 8, 9],
            "bar": ["1", "2", "3", "4", "5", "6", "7", "8", "9"],
            "date_month": [
                201001,
                201002,
                201003,
                201004,
                201005,
                201006,
                201007,
                201008,
                201009,
            ],
            "datetime": [
                datetime.fromisoformat("2010-01-01"),
                datetime.fromisoformat("2010-02-01"),
                datetime.fromisoformat("2010-03-01"),
                datetime.fromisoformat("2010-04-01"),
                datetime.fromisoformat("2010-05-01"),
                datetime.fromisoformat("2010-06-01"),
                datetime.fromisoformat("2010-07-01"),
                datetime.fromisoformat("2010-08-01"),
                datetime.fromisoformat("2010-09-01"),
            ],
            "static_part": ["A", "A", "A", "B", "B", "B", "C", "C", "C"],
        }
    )


def test_roundtrip_read(tmp_path, data_batch_1: pl.DataFrame):
    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
    )

    result = polars_scan_delta(tmp_path).collect()

    assert_frame_equal(result, data_batch_1)

    write_deltalake(tmp_path, data_batch_1.to_arrow(), mode="append")
    result = polars_scan_delta(tmp_path).collect()
    assert_frame_equal(result, pl.concat([data_batch_1] * 2))


def test_roundtrip_read_filter(tmp_path, data_batch_1: pl.DataFrame):
    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
    )

    result = polars_scan_delta(tmp_path).filter(pl.col("foo") > 5).collect()

    assert_frame_equal(result, data_batch_1.filter(pl.col("foo") > 5))

    write_deltalake(tmp_path, data_batch_1.to_arrow(), mode="append")
    result = polars_scan_delta(tmp_path).filter(pl.col("foo") > 5).collect()
    assert_frame_equal(result, pl.concat([data_batch_1] * 2).filter(pl.col("foo") > 5))


def test_roundtrip_read_partitioned(tmp_path, data_batch_1: pl.DataFrame):
    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
        partition_by=["date_month", "static_part"],
    )

    result = polars_scan_delta(tmp_path).collect()

    assert_frame_equal(result, data_batch_1, check_row_order=False)

    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
    )
    result = polars_scan_delta(tmp_path).collect()
    assert_frame_equal(result, pl.concat([data_batch_1] * 2), check_row_order=False)


def test_roundtrip_read_partitioned_filtered(tmp_path, data_batch_1: pl.DataFrame):
    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
        partition_by=["date_month", "static_part"],
    )

    result = (
        polars_scan_delta(tmp_path)
        .filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        )
        .collect()
    )

    assert_frame_equal(
        result,
        data_batch_1.filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        ),
        check_row_order=False,
    )

    write_deltalake(tmp_path, data_batch_1.to_arrow(), mode="append")
    result = (
        polars_scan_delta(tmp_path)
        .filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        )
        .collect()
    )
    assert_frame_equal(
        result,
        pl.concat([data_batch_1] * 2).filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        ),
        check_row_order=False,
    )


def test_roundtrip_read_partitioned_filtered_select(
    tmp_path, data_batch_1: pl.DataFrame
):
    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
        partition_by=["date_month", "static_part"],
    )

    result = (
        polars_scan_delta(tmp_path)
        .filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        )
        .select("foo")
        .collect()
    )

    assert_frame_equal(
        result,
        data_batch_1.filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        ).select("foo"),
        check_row_order=False,
    )

    write_deltalake(
        tmp_path,
        data_batch_1.to_arrow(),
        mode="append",
    )
    result = (
        polars_scan_delta(tmp_path)
        .filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        )
        .select("foo")
        .collect()
    )
    assert_frame_equal(
        result,
        pl.concat([data_batch_1] * 2)
        .filter(
            (pl.col("static_part") == "A")
            & (pl.col("date_month").is_in([201001, 201002]))
        )
        .select("foo"),
        check_row_order=False,
    )



from azure.core.credentials import TokenCredential, AccessToken
from datetime import datetime, timedelta


class FakeCredential(TokenCredential):
    def get_token(self, *args, **kwargs) -> AccessToken:
        return AccessToken(
            "fake_token", int((datetime.now() + timedelta(days=1)).timestamp())
        )


def test_get_storage_options_fsspec():
    from deltalake2db.azure_helper import get_storage_options_fsspec

    opt = get_storage_options_fsspec({"chain": "managed_identity"})
    assert opt["exclude_powershell_credential"] == True
    assert opt["exclude_managed_identity_credential"] == False
    assert opt["exclude_cli_credential"] == True


def test_get_storage_options_object_store():
    from deltalake2db.azure_helper import get_storage_options_object_store

    opts = [{"chain": "default"}, {"anon": False}]
    for o in opts:
        new_path, creds = get_storage_options_object_store(
            "az://accountnamethatslong.blob.core.windows.net/testlakedb/td/delta/fake",
            o,
            get_credential=lambda x: FakeCredential(),
        )
        assert isinstance(new_path, str)
        assert creds is not None
        assert ".blob.core" not in new_path
        assert ".dfs.core" not in new_path
        assert new_path == "az://testlakedb/td/delta/fake"
        assert creds["token"] == "fake_token"
        assert creds["account_name"] == "accountnamethatslong"
        new_path, creds = get_storage_options_object_store(
            "abfss://accountnamethatslong.dfs.core.windows.net/testlakedb/td/delta/fake",
            o,
            get_credential=lambda x: FakeCredential(),
        )
        assert isinstance(new_path, str)
        assert creds is not None
        assert ".blob.core" not in new_path
        assert ".dfs.core" not in new_path
        assert new_path == "abfss://testlakedb/td/delta/fake"
        assert creds["token"] == "fake_token"
        assert creds["account_name"] == "accountnamethatslong"



from collections import OrderedDict
import duckdb
import polars as pl
import pytest


@pytest.mark.skip(reason="Would need a real account for this")
def test_chain():
    from deltalake2db import duckdb_create_view_for_delta

    with duckdb.connect() as con:
        duckdb_create_view_for_delta(
            con,
            "az://testlakedb/td/delta/fake",
            "delta_table",
            storage_options={"use_emulator": "True", "chain": "default"},
        )

        df = pl.from_arrow(con.execute("select * from delta_table").fetch_arrow_table())
        assert isinstance(df, pl.DataFrame)
        print(df)
    main_coord_field = df.schema["main_coord"]
    assert isinstance(main_coord_field, pl.Struct)
    fields = main_coord_field.fields
    assert "lat" in [f.name for f in fields]
    assert "lon" in [f.name for f in fields]

    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    assert df.schema == OrderedDict(
        [
            ("Super Name", pl.String),
            ("Company Very Short", pl.String),
            ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
            ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
            ("age", pl.List(pl.Int64)),
            ("new_name", pl.String),
        ]
    )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


@pytest.mark.parametrize(
    "use_fsspec,use_delta_ext",
    [(True, False), (False, False), (False, True)],
)
def test_col_mapping(storage_options, use_fsspec: bool, use_delta_ext: bool):
    from deltalake2db import duckdb_create_view_for_delta

    if (
        use_delta_ext
    ):  # see https://github.com/duckdb/duckdb-delta/issues/50#issuecomment-2730119786
        select = ["Super Name", "Company Very Short", "age", "new_name"]
    else:
        select = None
    with duckdb.connect() as con:
        duckdb_create_view_for_delta(
            con,
            "az://testlakedb/td/delta/fake",
            "delta_table",
            storage_options=storage_options,
            use_fsspec=use_fsspec,
            use_delta_ext=use_delta_ext,
            select=select,
        )
        duckdb_create_view_for_delta(
            con,
            "az://testlakedb/td/delta/fake",
            "delta_table",
            storage_options=storage_options,
            use_fsspec=use_fsspec,
            use_delta_ext=use_delta_ext,
            select=select,
        )  # do it twice to test duplicate secrets

        df = pl.from_arrow(con.execute("select * from delta_table").fetch_arrow_table())
        assert isinstance(df, pl.DataFrame)
        print(df)
    if not use_delta_ext:
        main_coord_field = df.schema["main_coord"]
        assert isinstance(main_coord_field, pl.Struct)
        fields = main_coord_field.fields
        assert "lat" in [f.name for f in fields]
        assert "lon" in [f.name for f in fields]

    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    if not use_delta_ext:
        assert df.schema == OrderedDict(
            [
                ("Super Name", pl.String),
                ("Company Very Short", pl.String),
                ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
                ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
                ("age", pl.List(pl.Int64)),
                ("new_name", pl.String),
            ]
        )
    else:
        assert df.schema == OrderedDict(
            [
                ("Super Name", pl.String),
                ("Company Very Short", pl.String),
                ("age", pl.List(pl.Int64)),
                ("new_name", pl.String),
            ]
        )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


def test_empty_struct(storage_options):
    # >>> duckdb.execute("""Select { 'lat': 1 } as tester union all select Null""").fetchall()
    import pyarrow.compute as pc

    from deltalake2db import get_sql_for_delta

    with duckdb.connect() as con:
        sql = get_sql_for_delta(
            "az://testlakedb/td/delta/fake",
            duck_con=con,
            storage_options=storage_options,
        )
        print(sql)
        con.execute("create view delta_table as " + sql)

        df = con.execute("select * from delta_table").fetch_arrow_table()
        print(df)
        mc = (
            df.filter(pc.field("new_name") == "Hans Heiri")
            .select(["main_coord"])
            .to_pylist()
        )
        assert len(mc) == 1
        assert mc[0]["main_coord"] is None



from typing import (
    Sequence,
    Union,
    Protocol,
    TYPE_CHECKING,
    TypedDict,
    Literal,
    Optional,
)
import json

if TYPE_CHECKING:
    import duckdb

PrimitiveType = Literal[
    "string",
    "integer",
    "long",
    "float",
    "double",
    "boolean",
    "binary",
    "date",
    "timestamp",
    "timestamp without time zone",
    "decimal",
]


class PrimitiveField(TypedDict):
    name: str
    type: PrimitiveType


class MapType(TypedDict):
    type: Literal["map"]
    keyType: PrimitiveType
    valueType: "Union[StructType, PrimitiveType, ArrayType, MapType]"
    valueContainsNull: bool


class MapField(MapType):
    name: str


class ArrayType(TypedDict):
    type: Literal["array"]
    elementType: "Union[StructType, PrimitiveType, ArrayType, MapType]"


class ArrayField(ArrayType):
    name: str


Field = Union["StructField", "PrimitiveField", "ArrayField", "MapField"]


class StructType(TypedDict):
    type: Literal["struct"]
    fields: "Sequence[Field]"


class StructField(StructType):
    name: str


DataType = Union[StructType, PrimitiveType, ArrayType, MapType]


def is_primitive_type(t: DataType) -> bool:
    return isinstance(t, str)


class DeltaProtocol(TypedDict):
    minReaderVersion: int
    minWriterVersion: int
    readerFeatures: Optional[Sequence[str]]
    writerFeatures: Optional[Sequence[str]]


class MetaState:
    last_metadata: Union[dict, None] = None
    protocol: Union[DeltaProtocol, None] = None
    add_actions: dict[str, dict] = {}
    last_commit_info: Union[dict, None] = None

    @property
    def last_write_time(self):
        from datetime import datetime, timezone

        assert self.last_commit_info is not None
        ts = self.last_commit_info.get("timestamp", None)
        assert ts is not None
        return datetime.fromtimestamp(ts / 1000.0, timezone.utc)

    @property
    def schema(self) -> Union[StructType, None]:
        if self.last_metadata:
            sc = self.last_metadata.get("schemaString", None)
            if sc:
                return json.loads(sc)

        return None

    def __init__(self) -> None:
        self.last_metadata = None
        self.add_actions = {}


def process_meta_data(actions: dict, state: MetaState):
    if actions.get("metaData"):
        state.last_metadata = actions["metaData"]
    if actions.get("protocol"):
        state.protocol = actions["protocol"]
    if actions.get("add"):
        path = actions["add"]["path"]
        state.add_actions[path] = actions["add"]
    if actions.get("commitInfo"):
        state.last_commit_info = actions["commitInfo"]
    if actions.get("remove"):
        path = actions["remove"]["path"]
        if path in state.add_actions:
            del state.add_actions[path]


class MetadataEngine(Protocol):
    def read_jsonl(self, path: str) -> Sequence[dict]: ...

    def read_parquet(self, path: str) -> Sequence[dict]: ...


class PolarsEngine(MetadataEngine):
    def __init__(self, storage_options: Optional[dict]) -> None:
        super().__init__()
        self.storage_options = storage_options

    def read_jsonl(self, path: str) -> Sequence[dict]:
        try:
            import polars as pl

            df = pl.read_ndjson(path, storage_options=self.storage_options)
            return df.to_dicts()
        except OSError as e:
            if "404" in str(e) or "No such file or directory" in str(e):
                raise FileNotFoundError from e
            raise

    def read_parquet(self, path: str) -> Sequence[dict]:
        try:
            import polars as pl

            df = pl.read_parquet(path, storage_options=self.storage_options)
            return df.to_dicts()
        except OSError as e:
            if "BlobNotFound" in str(e):
                raise FileNotFoundError from e
            if "404" in str(e) or "No such file or directory" in str(e):
                raise FileNotFoundError from e
            raise


class DuckDBEngine(MetadataEngine):
    def __init__(self, con: "duckdb.DuckDBPyConnection") -> None:
        super().__init__()
        self.con = con

    def read_jsonl(self, path: str) -> Sequence[dict]:
        import duckdb

        q = f"SELECT * FROM read_json('{path}', format='newline_delimited')"
        with self.con.cursor() as cur:
            try:
                cur.execute(q)
                assert cur.description is not None
                desc = [d[0] for d in cur.description]
                return [dict(zip(desc, row)) for row in cur.fetchall()]
            except duckdb.IOException as e:
                if "BlobNotFound" in str(e):
                    raise FileNotFoundError from e
                if "No files found" in str(e):
                    raise FileNotFoundError from e
                raise

    def read_parquet(self, path: str) -> Sequence[dict]:
        import duckdb

        q = f"SELECT * FROM read_parquet('{path}')"
        with self.con.cursor() as cur:
            try:
                cur.execute(q)
                assert cur.description is not None
                desc = [d[0] for d in cur.description]
                return [dict(zip(desc, row)) for row in cur.fetchall()]
            except duckdb.IOException as e:
                if "No files found" in str(e):
                    raise FileNotFoundError from e
                raise


def _delta_fn(version: int) -> str:
    return f"{version:020d}"


def field_to_type(field: Field) -> DataType:
    if field["type"] == "struct":
        return field
    elif field["type"] == "array":
        return field
    elif field["type"] == "map":
        return field
    else:
        return field["type"]


def get_meta(
    engine: MetadataEngine, delta_path: str, *, version: Optional[int] = None
) -> MetaState:
    try:
        checkpoint = engine.read_jsonl(
            delta_path.rstrip("/") + "/_delta_log/_last_checkpoint"
        )[0]
    except FileNotFoundError:
        checkpoint = None
    state = MetaState()
    if checkpoint:
        check_point_version = checkpoint.get("version", 0)
        if version is not None and version < check_point_version:
            start_version = 0
        else:
            check_point_file = (
                delta_path.rstrip("/")
                + f"/_delta_log/{_delta_fn(check_point_version)}.checkpoint.parquet"
            )
            check_point_data = engine.read_parquet(check_point_file)
            for action in check_point_data:
                process_meta_data(action, state)
            start_version = check_point_version + 1
    else:
        start_version = 0
    current_version = start_version
    while version is None or current_version <= version:
        commit_file = (
            delta_path.rstrip("/") + f"/_delta_log/{_delta_fn(current_version)}.json"
        )
        try:
            commit_data = engine.read_jsonl(commit_file)
        except FileNotFoundError:
            break
        for action in commit_data:
            process_meta_data(action, state)
        current_version += 1
    return state



from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Optional, Union
import sqlglot.expressions as ex
from deltalake2db.azure_helper import (
    get_storage_options_fsspec,
    get_storage_options_object_store,
    get_account_name_from_path,
    AZURE_EMULATOR_CONNECTION_STRING,
)
from deltalake2db.filter_by_meta import _can_filter
import deltalake2db.sql_utils as squ
from deltalake2db.delta_meta_retrieval import (
    DataType,
    DuckDBEngine,
    get_meta,
    field_to_type,
    PrimitiveField,
    Field,
)

if TYPE_CHECKING:
    from azure.core.credentials import TokenCredential
    import duckdb


def _cast(s: ex.Expression, t: Optional[ex.DATA_TYPE]):
    if t is None:
        return s
    return ex.cast(s, t)


def _dummy_expr(
    field_type: "DataType | PrimitiveField",
) -> ex.Expression:
    if isinstance(field_type, str):
        if str(field_type).startswith("decimal("):
            cast_as = ex.DataType.build(str(field_type))
        else:
            cast_as = type_map.get(field_type)
        return _cast(ex.Null(), cast_as)
    elif field_type["type"] == "struct":
        return squ.struct(
            {
                subfield["name"]: _dummy_expr(subfield)
                for subfield in field_type["fields"]
            }
        )
    elif field_type["type"] == "array":
        return ex.Array(expressions=[_dummy_expr(field_type["elementType"])])
    elif field_type["type"] == "map":
        return ex.MapFromEntries(
            this=ex.Array(
                expressions=[
                    ex.Tuple(
                        expressions=[
                            _dummy_expr(field_type["keyType"]),
                            _dummy_expr(field_type["valueType"]),
                        ]
                    )
                ]
            )
        )
    else:
        field_type = field_type["type"]
        if str(field_type).startswith("decimal("):
            cast_as = ex.DataType.build(str(field_type))
        else:
            cast_as = type_map.get(field_type)
        return _cast(ex.Null(), cast_as)


def _get_expr(
    base_expr: "ex.Expression|None",
    dtype: "DataType",
    meta: "Optional[Field]",
    alias=True,
    *,
    counter: int = 0,
) -> "ex.Expression":
    pn = (
        meta.get("metadata", {}).get("delta.columnMapping.physicalName", meta["name"])
        if meta
        else None
    )
    if base_expr is None:
        assert pn is not None
        base_expr = ex.Column(this=pn)
    if isinstance(dtype, dict) and dtype["type"] == "struct":
        struct_expr = (
            ex.case()
            .when(base_expr.is_(ex.Null()), ex.Null())
            .else_(
                squ.struct(
                    {
                        subfield["name"]: _get_expr(
                            ex.Dot(
                                this=base_expr,
                                expression=ex.Identifier(
                                    this=subfield.get("metadata", {}).get(
                                        "delta.columnMapping.physicalName",
                                        subfield["name"],
                                    ),
                                    quoted=True,
                                ),
                            ),
                            field_to_type(subfield),
                            subfield,
                            alias=False,
                            counter=counter + 1,
                        )
                        for subfield in dtype["fields"]
                    }
                )
            )
        )
        if alias and meta is not None:
            return struct_expr.as_(meta["name"])
        return struct_expr
    elif isinstance(dtype, dict) and dtype["type"] == "array":
        vl = squ.list_transform(
            "x_" + str(counter),
            base_expr,
            _get_expr(
                ex.Identifier(this="x_" + str(counter), quoted=False),
                dtype["elementType"],
                meta=meta,
                alias=False,
                counter=counter + 1,
            ),
        )
        if alias and meta is not None:
            return vl.as_(meta["name"])
        return vl

    return base_expr.as_(meta["name"]) if meta is not None and alias else base_expr


def load_install_extension(con: "duckdb.DuckDBPyConnection", ext_name: str):
    with con.cursor() as cur:
        cur.execute(
            "select loaded, installed from duckdb_extensions() where extension_name=$ext_name ",
            {
                "ext_name": ext_name,
            },
        )
        res = cur.fetchone()
        loaded = res[0] if res else False
        installed = res[0] if res else False
    if not installed:
        con.install_extension(ext_name)
    if not loaded:
        con.load_extension(ext_name)


def apply_storage_options(
    con: "duckdb.DuckDBPyConnection",
    uri: Union[str, Path],
    storage_options: Optional[dict],
    *,
    use_fsspec: bool = False,
):
    if storage_options is None:
        return
    if isinstance(uri, str) and (
        ".blob.core.windows.net" in uri or ".dfs.core.windows.net" in uri
    ):
        from urllib.parse import urlparse

        up = urlparse(uri)
        account_name_from_url = up.netloc.split(".")[0]
    else:
        account_name_from_url = None
    if use_fsspec:
        apply_storage_options_fsspec(
            con,
            uri if isinstance(uri, str) else str(uri.absolute()),
            storage_options,
            account_name_from_url,
        )
    else:
        apply_storage_options_azure_ext(
            con,
            storage_options,
            account_name_path=account_name_from_url,
        )


def apply_storage_options_fsspec(
    con: "duckdb.DuckDBPyConnection",
    base_path: str,
    storage_options: dict,
    account_name_path: Optional[str] = None,
):
    use_emulator = storage_options.get("use_emulator", "0") in ["1", "True", "true"]

    opts = get_storage_options_fsspec(storage_options)
    account_name = storage_options.get(
        "account_name",
        storage_options.get("azure_storage_account_name", account_name_path),
    )
    if not use_emulator:
        assert account_name is not None, "account_name must be provided for fsspec"
    proto_name_duckdb = "fsspec_az_" + (account_name or "_emulator")
    if not con.filesystem_is_registered(proto_name_duckdb):
        from .fsspec_fs import get_az_blob_fs

        fs = get_az_blob_fs(proto_name_duckdb, **opts)

        con.register_filesystem(fs)  # type: ignore

    return proto_name_duckdb


def apply_storage_options_azure_ext(
    con: "duckdb.DuckDBPyConnection",
    storage_options: dict,
    *,
    type="azure",
    account_name_path: Optional[str] = None,
):
    if type in ["az", "abfs", "abfss"]:
        type = "azure"
    if type != "azure":
        raise ValueError("Only azure is supported for now")
    load_install_extension(con, "azure")
    with con.cursor() as cur:
        cur.execute("FROM duckdb_secrets() where type='azure';")
        secrets = cur.fetchall()
        se_names = [s[0] for s in secrets]
    if len(se_names) > 0:  # for now, only one secret is supported in DuckDB, it seems
        return
    account_name = storage_options.get(
        "account_name",
        storage_options.get("azure_storage_account_name", account_name_path),
    )
    if str(storage_options.get("anon", "0")).lower() in ["1", "true"]:
        return None  # no need to create a secret for anon (which is an fsspec-thing)
    use_emulator = storage_options.get("use_emulator", "0") in ["1", "True", "true"]
    secret_name = account_name or ("_emulator" if use_emulator else "")

    if secret_name not in se_names:
        if "connection_string" in storage_options:
            cr = storage_options["connection_string"]
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{cr}'
);"""
            )
        elif account_name is not None and "account_key" in storage_options:
            ak = storage_options["account_key"]
            conn_str = f"AccountName={account_name};AccountKey={ak};BlobEndpoint=https://{account_name}.blob.core.windows.net;"
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{conn_str}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None and "sas_token" in storage_options:
            sas_token = storage_options["sas_token"]
            conn_str = f"BlobEndpoint=https://{account_name}.blob.core.windows.net;SharedAccessSignature={sas_token}"
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{conn_str}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None and "client_id" in storage_options:
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
PROVIDER SERVICE_PRINCIPAL,
TENANT_ID  '{storage_options.get("tenant_id")}',
CLIENT_ID   '{storage_options.get("client_id")}',
CLIENT_SECRET    '{storage_options.get("client_secret")}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif account_name is not None:
            chain = storage_options.get("chain", "default")
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
PROVIDER CREDENTIAL_CHAIN,
CHAIN '{chain}',
ACCOUNT_NAME '{account_name}'
);"""
            )
        elif use_emulator:
            con.execute(
                f"""CREATE SECRET {secret_name} (
TYPE AZURE,
CONNECTION_STRING '{AZURE_EMULATOR_CONNECTION_STRING}',
ACCOUNT_NAME 'devstoreaccount1'
);"""
            )
        else:
            raise ValueError(
                "No connection string or account_name and account_key provided"
            )
    return secret_name


type_map = {
    "byte": ex.DataType.Type.TINYINT,
    "int": ex.DataType.Type.INT,
    "integer": ex.DataType.Type.INT,
    "long": ex.DataType.Type.BIGINT,
    "boolean": ex.DataType.Type.BOOLEAN,
    "date": ex.DataType.Type.DATE,
    "timestamp": ex.DataType.Type.TIMESTAMPTZ,
    "float": ex.DataType.Type.FLOAT,
    "double": ex.DataType.Type.DOUBLE,
    "string": ex.DataType.Type.VARCHAR,
    "short": ex.DataType.Type.SMALLINT,
    "binary": ex.DataType.Type.BINARY,
    "timestampNtz": ex.DataType.Type.TIMESTAMP,
    "timestamp_ntz": ex.DataType.Type.TIMESTAMP,
    "decimal": ex.DataType.Type.DECIMAL,
}


def create_view_for_delta(
    con: "duckdb.DuckDBPyConnection",
    delta_table: "Union[Path , str]",
    view_name: str,
    overwrite=True,
    *,
    conditions: Optional[dict] = None,
    storage_options: Optional[dict] = None,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec: bool = False,
    use_delta_ext=False,
    select: "Optional[list[str]]" = None,
    version: "Optional[int]" = None,
):
    sql = get_sql_for_delta(
        delta_table,
        duck_con=con,
        conditions=conditions,
        storage_options=storage_options,
        get_credential=get_credential,
        use_fsspec=use_fsspec,
        use_delta_ext=use_delta_ext,
        select=select,
        version=version,
    )
    assert '"' not in view_name
    if overwrite:
        con.execute(f'create or replace view "{view_name}" as {sql}')
    else:
        con.execute(f'create view "{view_name}" as {sql}')


def get_sql_for_delta_expr(
    table_or_path: "Union[Path , str]",
    conditions: Union[Optional[dict], Sequence[ex.Expression], ex.Expression] = None,
    select: Union[Sequence[Union[str, ex.Expression]], None] = None,
    distinct=False,
    cte_wrap_name: Union[str, None] = None,
    action_filter: Union[Callable[[dict], bool], None] = None,
    sql_prefix="delta",
    delta_table_cte_name: Union[str, None] = None,
    duck_con: "Union[duckdb.DuckDBPyConnection, None]" = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec=False,
    use_delta_ext=False,
    version: "Optional[int]" = None,
) -> ex.Select:
    from .sql_utils import read_parquet, union, filter_via_dict

    if isinstance(table_or_path, str):
        base_path = table_or_path
    else:
        base_path = str(table_or_path.absolute())

    base_path = base_path.removesuffix("/")
    is_azure = base_path.startswith("az://") or base_path.startswith("abfss://")
    account_name_path = get_account_name_from_path(base_path) if is_azure else None

    conds = filter_via_dict(conditions) if isinstance(conditions, dict) else conditions
    owns_con = False
    if duck_con is None:
        import duckdb

        duck_con = duckdb.connect()
        owns_con = True
    if use_delta_ext:
        load_install_extension(duck_con, "delta")
    if use_fsspec:
        fake_protocol = apply_storage_options_fsspec(
            duck_con,
            base_path,
            storage_options or {},
            account_name_path=account_name_path,
        )
        base_path = fake_protocol + "://" + base_path.split("://")[1]
    elif is_azure:
        apply_storage_options_azure_ext(
            duck_con,
            storage_options or {},
            type="azure",
            account_name_path=account_name_path,
        )  # type: ignore

    try:
        if not use_delta_ext:
            from .protocol_check import check_is_supported

            meta_state = get_meta(DuckDBEngine(duck_con), base_path, version=version)
            check_is_supported(meta_state)
            delta_table_cte_name = delta_table_cte_name or sql_prefix + "_delta_table"

            file_selects: list[ex.Select] = []
            assert meta_state.schema is not None
            delta_fields = meta_state.schema["fields"]
            for ac in meta_state.add_actions.values():
                if (
                    conditions is not None
                    and isinstance(conditions, dict)
                    and _can_filter(ac, conditions)
                ):
                    continue
                if action_filter and not action_filter(ac):
                    continue
                fullpath = base_path + "/" + ac["path"]
                with duck_con.cursor() as cur:
                    cur.execute(f"select name from parquet_schema('{fullpath}')")
                    cols: list[str] = [c[0] for c in cur.fetchall()]
                cols_sql: list[ex.Expression] = []
                for field in delta_fields:
                    field_name = field["name"]
                    phys_name = field.get("metadata", {}).get(
                        "delta.columnMapping.physicalName", field_name
                    )

                    cast_as = None

                    if isinstance(field["type"], str):
                        if str(field["type"]).startswith("decimal("):
                            cast_as = ex.DataType.build(str(field["type"]))
                        else:
                            cast_as = type_map.get(field["type"])
                    if "partitionValues" in ac and phys_name in ac["partitionValues"]:
                        cols_sql.append(
                            _cast(
                                ex.convert(ac["partitionValues"][phys_name]), cast_as
                            ).as_(field_name)
                        )
                    elif "partition." + phys_name in ac:
                        cols_sql.append(
                            _cast(
                                ex.convert(ac["partition." + phys_name]), cast_as
                            ).as_(field_name)
                        )
                    elif (
                        "partitionValues" in ac
                        and field["name"] in ac["partitionValues"]
                    ):
                        cols_sql.append(
                            _cast(
                                ex.convert(ac["partitionValues"][field["name"]]),
                                cast_as,
                            ).as_(field_name)
                        )
                    elif "partition." + field["name"] in ac:
                        cols_sql.append(
                            _cast(
                                ex.convert(ac["partition." + field["name"]]), cast_as
                            ).as_(field_name)
                        )
                    elif phys_name in cols:
                        cols_sql.append(
                            _get_expr(
                                ex.column(phys_name, quoted=True),
                                field_to_type(field),
                                field,
                            )
                        )
                    else:
                        cols_sql.append(ex.Null().as_(field_name))

                select_pq = ex.select(
                    *cols_sql
                ).from_(
                    read_parquet(ex.convert(fullpath))
                )  # "SELECT " + ", ".join(cols_sql) + " FROM read_parquet('" + fullpath + "')"
                file_selects.append(select_pq)
            if len(file_selects) == 0:
                file_selects = []
                fields = [
                    _dummy_expr(field).as_(field["name"]) for field in delta_fields
                ]
                file_selects.append(ex.select(*fields).where("1=0"))
            file_sql = ex.CTE(
                this=union(file_selects, distinct=False), alias=delta_table_cte_name
            )
            if select:
                select_exprs = [
                    ex.column(s, quoted=True) if isinstance(s, str) else s
                    for s in select
                ]
            else:
                select_exprs = [ex.Star()]

            se = ex.select(*select_exprs)
            if distinct:
                se = se.distinct()
            if conds is not None:
                se = se.where(*conds)
            se = se.from_(delta_table_cte_name)

            if cte_wrap_name:
                s = ex.select(ex.Star()).from_(cte_wrap_name)
                # s = s.with_()
                s.with_(file_sql.alias, file_sql.args["this"], copy=False)
                s.with_(cte_wrap_name, se, copy=False)
                return s
            else:
                se.with_(file_sql.alias, file_sql.args["this"], copy=False)
                return se
        else:
            assert version is None, "version is not supported with delta extension"
            if select:
                select_exprs = [
                    ex.column(s, quoted=True) if isinstance(s, str) else s
                    for s in select
                ]
            else:
                select_exprs = [ex.Star()]

            se = ex.select(*select_exprs).from_(
                ex.func("delta_scan", ex.convert(base_path))
            )
            if distinct:
                se = se.distinct()
            if conds is not None:
                se = se.where(*conds)
            return se

    finally:
        if owns_con:
            duck_con.close()


def get_sql_for_delta(
    dt: "Union[Path, str]",
    conditions: Optional[dict] = None,
    select: Union[list[str], None] = None,
    distinct=False,
    action_filter: Union[Callable[[dict], bool], None] = None,
    cte_wrap_name: Union[str, None] = None,
    sql_prefix="delta",
    duck_con: "Union[duckdb.DuckDBPyConnection, None]" = None,
    storage_options: Optional[dict] = None,
    *,
    get_credential: "Optional[Callable[[str], Optional[TokenCredential]]]" = None,
    use_fsspec: bool = False,
    use_delta_ext=False,
    version: "Optional[int]" = None,
) -> str:
    expr = get_sql_for_delta_expr(
        table_or_path=dt,
        conditions=conditions,
        select=select,
        distinct=distinct,
        action_filter=action_filter,
        sql_prefix=sql_prefix,
        cte_wrap_name=cte_wrap_name,
        duck_con=duck_con,
        storage_options=storage_options,
        get_credential=get_credential,
        use_fsspec=use_fsspec,
        use_delta_ext=use_delta_ext,
        version=version,
    )
    if cte_wrap_name:
        suffix_sql = ex.select(ex.Star()).from_(cte_wrap_name).sql(dialect="duckdb")
        ws = expr.sql(dialect="duckdb")
        without_suffix = ws.removesuffix(suffix_sql)
        assert without_suffix != ws
        return without_suffix
    return expr.sql(dialect="duckdb")



from collections import OrderedDict
from typing import Union
from deltalake import DeltaTable
import polars as pl
import pytest


@pytest.mark.parametrize("use_pyarrow", [True, False])
def test_col_mapping(use_pyarrow):
    dt = "tests/data/faker2"

    from deltalake2db import polars_scan_delta, PolarsSettings

    df = polars_scan_delta(dt, settings=PolarsSettings(use_pyarrow=use_pyarrow))

    df = df.collect() if not isinstance(df, pl.DataFrame) else df

    coord_field = df.schema["main_coord"]
    assert isinstance(coord_field, pl.Struct)
    fields = coord_field.fields
    assert "lat" in [f.name for f in fields]
    assert "lon" in [f.name for f in fields]
    age_field = df.schema["age"]
    assert isinstance(age_field, pl.List)
    assert isinstance(age_field.inner, pl.Int64)
    assert df.schema == OrderedDict(
        [
            ("Super Name", pl.String),
            ("Company Very Short", pl.String),
            ("main_coord", pl.Struct({"lat": pl.Float64, "lon": pl.Float64})),
            ("coords", pl.List(pl.Struct({"lat": pl.Float64, "lon": pl.Float64}))),
            ("age", pl.List(pl.Int64)),
            ("new_name", pl.String),
        ]
    )

    as_py_rows = df.rows(named=True)
    print(as_py_rows)


def _collect(df: Union[pl.DataFrame, pl.LazyFrame]):
    return df.collect() if not isinstance(df, pl.DataFrame) else df


@pytest.mark.parametrize("use_pyarrow", [True, False])
def test_user_add(use_pyarrow):
    import shutil
    import pyarrow as pa

    shutil.rmtree("tests/data/_user3", ignore_errors=True)
    shutil.copytree("tests/data/user", "tests/data/_user3")
    dt = DeltaTable("tests/data/_user3")

    old_version = dt.version()
    from deltalake.writer import write_deltalake
    from deltalake import __version__ as deltalake_version

    if deltalake_version.startswith("0."):
        engine_args = {"engine": "rust"}
    else:
        engine_args = {}
    data = [
        {
            "User - iD": 1555,
            "FirstName": "Hansueli",
            "metadata": {"key1": "value1", "key2": "value2"},
        }
    ]
    schema = pa.schema(
        [
            pa.field("User - iD", pa.int64()),
            pa.field("FirstName", pa.string()),
            pa.field(
                "metadata",
                pa.map_(pa.string(), pa.string()),
            ),
        ]
    )
    table = pa.Table.from_pylist(data, schema=schema)

    write_deltalake(
        dt,
        table,
        schema_mode="overwrite",
        mode="overwrite",
        **engine_args,  # type: ignore
    )
    dt.update_incremental()

    dt_o = DeltaTable("tests/data/_user3")
    dt_o.load_as_version(old_version)

    from deltalake2db import polars_scan_delta, PolarsSettings
    import polars as pl

    nc = _collect(
        polars_scan_delta(
            dt.table_uri, settings=PolarsSettings(use_pyarrow=use_pyarrow)
        ).select(pl.col("User - iD"))
    ).to_dicts()
    oc = _collect(
        polars_scan_delta(
            dt_o.table_uri,
            settings=PolarsSettings(use_pyarrow=use_pyarrow),
            version=old_version,
        ).select(pl.col("User - iD"))
    ).to_dicts()
    diff = [o["User - iD"] for o in nc if o not in oc]
    assert diff == [1555]


def test_user_empty():
    dt = "tests/data/user_empty"

    from deltalake2db import polars_scan_delta

    df = _collect(polars_scan_delta(dt))
    assert df.shape[0] == 0
    assert "time stämp" in df.columns


def test_select():
    dt = "tests/data/user"

    from deltalake2db import polars_scan_delta, PolarsSettings

    df = _collect(polars_scan_delta(dt, settings=PolarsSettings(fields=["User - iD"])))
    assert len(df.columns) == 1
    assert "User - iD" in df.columns

    df = _collect(
        polars_scan_delta(dt, settings=PolarsSettings(exclude_fields=["User - iD"]))
    )
    assert len(df.columns) > 1
    assert "User - iD" not in df.columns


@pytest.mark.parametrize("use_pyarrow", [True, False])
def test_strange_cols(use_pyarrow):
    dt = "tests/data/user"

    from deltalake2db import polars_scan_delta, PolarsSettings

    df = polars_scan_delta(dt, settings=PolarsSettings(use_pyarrow=use_pyarrow))

    df = _collect(df)
    col_names = df.columns
    assert "time stämp" in col_names


@pytest.mark.parametrize("use_pyarrow", [True, False])
def test_filter_number(use_pyarrow):
    dt = "tests/data/user"

    from deltalake2db import polars_scan_delta, PolarsSettings

    df = polars_scan_delta(
        dt, conditions={"Age": 23.0}, settings=PolarsSettings(use_pyarrow=use_pyarrow)
    )
    res = _collect(df).to_dicts()
    assert len(res) == 1
    assert res[0]["FirstName"] == "Peter"

    df2 = polars_scan_delta(dt, conditions={"Age": 500})

    assert df.schema == df2.schema, "Schema does not match"


def test_filter_name():
    dt = "tests/data/user"

    from deltalake2db import polars_scan_delta

    df = polars_scan_delta(dt, conditions={"FirstName": "Peter"})
    res = _collect(df).to_dicts()
    assert len(res) == 1
    assert res[0]["FirstName"] == "Peter"


def test_schema():
    from deltalake2db import polars_scan_delta, get_polars_schema

    for tbl in ["user", "faker2", "user_empty"]:
        dt = "tests/data/" + tbl

        df = polars_scan_delta(dt)
        schema = get_polars_schema(dt)

        assert df.schema == schema, f"Schema for {tbl} does not match"


def test_empty_struct():
    dt = "tests/data/faker2"

    from deltalake2db import polars_scan_delta

    df = polars_scan_delta(dt)

    df = _collect(df)

    mc = df.filter(new_name="Hans Heiri").select("main_coord").to_dicts()
    assert len(mc) == 1
    assert mc[0]["main_coord"] is None
